(module info setup/infotab (#%module-begin (define package-content-state (quote (built "9.1"))) (define collection (quote multi)) (define version "9.1") (define deps (quasiquote ("racket-lib" ("racket" #:version (unquote version))))) (define implies (quote (core))) (define pkg-desc "Racket libraries that are currently always available") (define pkg-authors (quote (mflatt))) (define license (quote (Apache-2.0 OR MIT)))))
