#lang info

(define release-note-files '(("Racket core" "HISTORY.txt"
                              -100
                              (("Porting from v4.x to v5.x" "Racket_5.txt")
                               ("Porting from v3xxx to v4.x" "MzScheme_4.txt")
                               ("Porting from v2xxx to v3xxx" "MzScheme_300.txt")
                               ("Porting from v1xxx to v2xxx" "MzScheme_200.txt")))))

(define copy-man-pages '("racket.1"))

(define install-collection "private/tethered-installer.rkt")
